/*
 * pipe.c
 *
 * functions for reading/writing compressed files or through pipes
 *    - MY_OPEN: bzopen, gzopen, popen, fopen, mmap, open
 *          
 *   ToDo: setvbuf could reduce syscalls to read (slow)
 *     but check that setvbuf works system dependend with strace (utils/vbuf.c)
 *     if so, we better should use primitiv read, write and own_cache (stdlib)
 *     but what about popen? (stdio) replace by pipe and fork/exec/system
 *       also read can be interupted (O_NONBLOCK, ret<len, select()),   
 *     popen may be faster on multiproc-PC because of async i/o and read before open (just after close),
 *      but check that theory by explicit timing/debugging incl. iostat
 *     may be popen should be replaced by own_mpi=(mpi or pipe or thread or seq)
 *     create MY_FILE for extern calls
 *     fwrite/write 2GB-limits?
 *
 * history:
 *   v0.5: - fork makes a (virtual) copy of the parent process (bigmem*2),
 *           which will be released by exec (ToDo: check with tracer!),
 *           this can be fail on some systems if memory < 2*used_mem
 *           solution(?): fork at before mallocs and fork in child on cmd ???
 *   v0.6 - popen(gzip) => same problem
 *          starting gzip (which needs only some kB) from spin (which
 *           has GB allocated) needs TEMPORARLY GB's for internal fork on IRIX64
 *          perror=Resource temporarily unavailable
 *   v0.7 - STOREH=2 one early fork, malloc, fork by child via commands like
 *   v0.9 - da STOREH=2 auf SGI-THOR bei >2GB >100h pg mit fehlermeldung stoppt
 *          STOREH=3 using zlib.h
 *          STOREH=4 using raw-files (save)
 *   v0.10 - STOREH 0 und 2 entfernt um einfacheren 
 *   v0.11 - do not print perror if errno==0
 *   v2.26 - reduce B_NUM*B_NUM to F_NUM (avoid random access to disks)
 *           get maximum sequential speed
 *   v2.27 - memory only added STOREH=0, direct read/write STOREH=2
 *   ToDo:   F_NUM I/O threads + 2*B_NUM buffers for async I/O
 *           store to RAM-list (data+pointer-to-next)
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  /* sleep() defined */
#include <string.h>
#include "error.h"

#ifdef _AIX
#include <sys/select.h>
#include <strings.h>
#endif

/* #include "config.h" */ /* F_NUM file buffers! <= B_NUM */
#include "block.h"      /* B_NUM for 2*B_NUM async I/O buffers, F_NUM too */


#ifndef F_NUM  /* obsolete */
#error "please define F_NUM = {1..B_NUM} the number of H-files (I/O-threads)"
#endif
#if (F_NUM > B_NUM) || (F_NUM < 1) 
#error "F_NUM should be in the range 1 .. B_NUM"   
#endif

int pvv=0;	/* DEBUG-MODE parent-child-subchild communication */
const char *pipe_mode[5]={"realloc","pipe(gzip)","read","zlib","fread"};

/* obsolete?: 8M>>(F_NUM) buffer for fwrite/fread or 0 for unbuffered */
/* v2.29: makes no sence because we store Hblocks already
 *        but Tru64+SunOS may read single bytes if unbuffered (use STOREH=2!)
 */
#ifndef PBUFL
#define PBUFL (1024*1024*8) /* 8MB(Hsize multiple), 0 = no buffering */
#endif

/* forking-methods dont work on IRIX with big calculations 
 *   because fork temporary doubles memory consumtion, 3 should work
 * 1=popen(gzip), 3=zlib_gzread 4=fread 0=mem 2=read
 */
#ifndef STOREH
#if defined(HAVE_ZLIB)
#define STOREH 3     /* default using zlib if HAVE_ZLIB is defined */
#else 
#define STOREH 4     /* default pure read/write */
#endif
#endif

#if   STOREH == 0
/* ---- use malloc to get/store data ---- */ 
/* v2.29: realloc is bad for big arrays, because we need size_old+size_new */
/* therefore we use a chain, the read/write blocks must have fixed length */
int      fn[F_NUM]; /* file-descriptor replacement == BlockNo */
typedef struct mem_chain {
      char *buff;
      struct mem_chain *next;
      size_t len;
} t_mem_chain;
struct mem_chain  *mem[F_NUM]; /* start buffer */
struct mem_chain *mpos[F_NUM]; /* position pointer */

int mopen(int blk)  { mpos[blk]=mem[blk]; fn[blk]=blk; return(blk); }
int mclose(int blk) { mpos[blk]=0;                     return(blk); }

/* pointer is at last block, mixed r/w not allowed */
size_t mwrite(void *from, size_t len, int blk) {
  struct mem_chain *pp;
  /* pp is the pointer, where the data will be stored */
  if (!mem[blk]) { /* no chain root, once per run */
    mpos[blk]=mem[blk]=(struct mem_chain *)malloc(sizeof(struct mem_chain));
    if (!mem[blk]) return 0;
    mem[blk]->next=NULL;
    mem[blk]->buff=NULL;
    mem[blk]->len=0;
  }
  pp=mpos[blk];
  if (!pp->next) { /* empty frame for next step */
    pp->next=(struct mem_chain *)malloc(sizeof(struct mem_chain));
    if (!pp->next) return 0;
    pp->next->next=NULL;
    pp->next->buff=NULL;
    pp->next->len=0;
  }
  if (!pp->buff) {
    pp->buff=(char *)malloc(len);
    if (!pp->buff) return 0;
    pp->len=len;
  }
  if (pp->len!=len) {
    char *tmp;
    tmp=(char *)realloc(pp->buff, len);
    if (tmp) { pp->buff=tmp; pp->len=len; }
    else     { printf("\nERROR %d.%d pipe.c realloc %.0f failed\n",
                  mpi_i, blk, 1.*len);
                  return 0;
              }
    E0("DBG realloc, should not happen");
  }
  memcpy(pp->buff, from, len);
//  printf("\nDBG w %.0f %p",1.*len, pp);
  mpos[blk]=pp->next;
  return len;
}

size_t mread(void *ptr, size_t len, int blk) {
//  printf("\nDBG r %.0f %.0f %p",1.*len,1.*mpos[blk]->len,mpos[blk]);
  if (mpos[blk]==NULL) {F0("noMem");return 0;}
  if (mpos[blk]->len != len) {F0("blocklen read!=write");return 0;}
  memcpy(ptr, mpos[blk]->buff, len); mpos[blk]=mpos[blk]->next;
  return len;
}

#define SET_BUF             /* no buffer used */
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    mopen(blk)
#define ROPEN(name,blk)    mopen(blk)
#define CLOSE   mclose
#define ZERO    -1
#define READ(buf,len,fn)   mread( buf, len, fn)
#define WRITE(buf,len,fn)  mwrite( buf, len, fn)
/* once per run */
#define INI(blk) { fn[blk]=ZERO; mem[blk]=0; mpos[blk]=0; }

#elif   STOREH == 1
/* ---- use popen to read/write data ---- */
/* calls fork which can cause problems on systems doubling memory need */
/* #define FILETYPE FILE ToDo! */
FILE *fn[F_NUM];
#define SET_BUF
#define WOPENS "gzip -fcd %s"
#define ROPENS "gzip -fc1 >%s"
#define WOPEN(name,blk)    popen(name,"w")
#define ROPEN(name,blk)    popen(name,"r")
#define CLOSE   pclose
#define ZERO    (FILE *)0
#define READ(buf,len,fn)    fread( buf, 1, len, fn)
#define WRITE(buf,len,fn)  fwrite( buf, 1, len, fn)
#define INI(blk) fn[blk]=ZERO

#elif   STOREH == 2
/* ---- use read/write data ---- */ 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef O_NOATIME
#define O_NOATIME 0
#endif
#ifndef O_LARGEFILE    /* Tru64 */
#define O_LARGEFILE 0
#endif
int fn[F_NUM];  /* file handle */
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    open(name,O_WRONLY+O_CREAT+O_LARGEFILE+O_NOATIME,6*64+4*8+4)
#define ROPEN(name,blk)    open(name,O_RDONLY+O_LARGEFILE+O_NOATIME)
#define CLOSE   close
#define ZERO    0
#define READ(buf,len,fn)    read( fn, buf, len)
#define WRITE(buf,len,fn)  write( fn, buf, len)
#define INI(blk) fn[blk]=ZERO

#elif STOREH == 3  &&  defined(HAVE_ZLIB)
/* ---- use zlib or bzlib to read/write data ---- */
/* HAVE_ZLIB is defined in Makefile and given to gcc via -D macro=value */
#ifdef HAVE_BZLIB
  /* using bzlib (ca. 900kB of RAM used, bz2=1/3 unzip-time=4*gz) */
#include <bzlib.h>
BZFILE *fn[F_NUM];
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    BZ2_bzopen(name,"wb")
#define ROPEN(name,blk)    BZ2_bzopen(name,"rb")
#define CLOSE   BZ2_bzclose
#define ZERO    0
#define READ(buf,len,fn)   BZ2_bzread( fn, buf, len)
#define WRITE(buf,len,fn)  BZ2_bzwrite( fn, buf, len)
#define INI(blk) fn[blk]=ZERO
#else /* HAVE_BZLIB not defined */
  /* using zlib (ca. 260kB of dyn. RAM used, gz=1/2) */
#include <zlib.h>
gzFile fn[F_NUM];
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    gzopen(name,"wb2f")
#define ROPEN(name,blk)    gzopen(name,"rb")
#define CLOSE   gzclose
#define ZERO    0
#define READ(buf,len,fn)   gzread( fn, buf, len)
#define WRITE(buf,len,fn)  gzwrite( fn, buf, len)
#define INI(blk) fn[blk]=ZERO
#endif /* !BZLIB */

#elif STOREH == 4	/* std-file */
/* ---- default method using fread/fwrite data ---- */
FILE *fn[F_NUM];
#if PBUFL>0
char pbuf[F_NUM][PBUFL];
#define SET_BUF  setvbuf(fn[block],pbuf[block],_IOFBF,PBUFL)
#else
#define SET_BUF  setvbuf(fn[block],(char *)0,_IONBF,0)
#endif
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    fopen(name,"wb")
#define ROPEN(name,blk)    fopen(name,"rb")
#define CLOSE   fclose
#define ZERO    (FILE *)0
#if 0
#define READ(buf,len,fn)   fread( buf, 1, len, fn)
#define WRITE(buf,len,fn)  fwrite( buf, 1, len, fn)
#else
#define READ(buf,len,fn)  (fread( buf, len, 1, fn)*len)
#define WRITE(buf,len,fn) (fwrite( buf, len, 1, fn)*len)
#endif
#define INI(blk) fn[blk]=ZERO
/* see utils/vbuf.c for i/o-experience */
#else
# error "no such STOREH defined"
#endif

/* ------------------------------- close ----------------------- */
int pipe_close(int block){
 if (fn[block]!=ZERO) CLOSE(fn[block]); /* only close if open */
 fn[block]=ZERO;                  /* mark as closed */
 return 0;
}
/* ------------------------------- read ------------------------ */
int pipe_ropen(char *name,int block){
    char s[100];int i; sprintf(s,WOPENS,name);
    if (fn[block]!=ZERO) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=ZERO;i<5 && fn[block]==ZERO;i++){
      fn[block]=ROPEN(s,block);
      if (fn[block]==ZERO) { E2("pipe_ropen %s %d",s,i);perror(" "); 
       sleep(60); 
      }
    } if (fn[block]==ZERO)   E0("pipe_ropen failed");
    if( i>1 ) E0("pipe_ropen => ok");
    SET_BUF;
    return 0;
}
int pipe_read(char *buf,int len,int block){
    int nbytes=0; if(buf)buf[0]=0;
    if (fn[block]==ZERO) return 0;
    nbytes    = (int)READ( buf, len, fn[block]);
/*
    if (nbytes != len){ if(errno) perror("pipe_read");
      nbytes += (int)READ( buf+nbytes, len-nbytes, fn[block]);
      if (nbytes != len){ if(errno) perror("pipe_read"); }
    }
*/
    return nbytes;
}
/* --------------------------- write --------------------- */
int pipe_wopen(char *name,int block){
    char s[100];int i; sprintf(s,ROPENS,name);
    if (fn[block]!=ZERO) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=ZERO;i<5 && fn[block]==ZERO;i++){
      fn[block]=WOPEN(s,block);
      if (fn[block]==ZERO) { E2("pipe_wopen %s %d",s,i);perror(" ");
       sleep(30); }
    } 
    if (fn[block]==ZERO) { E0("pipe_wopen => failed"); return 1; }
    if( i>1 ) E0("pipe_wopen => ok");
    SET_BUF;
    return 0;
}
int pipe_write(char *s,int l,int block){
    int nbytes;
    if (fn[block]==ZERO) return 0;
    nbytes=(int)WRITE(s,l,fn[block]);
    if(nbytes != l) perror("pipe_write"); return nbytes;
}
int sub_ini(){
  int i;
  if (!mpi_i)
    printf(" pipe.c sub_ini STOREH=%d %s 8*sizeof(size_t)=%d\n",
            STOREH, pipe_mode[STOREH], 8*(int)sizeof(size_t));
  for(i=0;i<F_NUM;i++) INI(i);
  return 0;
}
int sub_end(){return 0;}
/* ================================================================= */

#if 0		/* test this */
int buf[1024*1024];
int main()
{       int i,j,k;char s[100+1];
	FILE *f1;
	setvbuf(stdout,(char *)NULL,_IONBF,0);
        sub_ini();				/* do this early!!! */
        s[100]=0;buf[0]=buf[1024*1023]=1;
#define FLEN 100 /* 100x18 */
        pipe_wopen("tmp000.gz");			/* gzip */
        for(i=1;i<FLEN;i++)
        pipe_write("Das ist ein Test.\n",18);
        pipe_close();
        printf("tmp000.gz written\n"); sleep(10);
        for(k=0;k<100000;k++){
          pipe_ropen("tmp000.gz");			/* gzip */
          for(i=1;i<FLEN;i++){
            j=pipe_read(s,18);s[j]=0;		/* timeout-zeit? */
            if(j!=18) fprintf(stderr," error pipe_read(%d)=%d\n",k,j);
            printf("%s",s);
          }
          pipe_close();
        }
        sub_end();
	return 0;
}
#endif
